from . import __path__
from .common import S3Artifact, jsonpickle
from .config import config
from .io import (InputArtifact, InputParameter, Inputs, OutputArtifact,
                 OutputParameter)
from .op_template import PythonScriptOPTemplate, ShellOPTemplate


class InitArtifactForSlices(PythonScriptOPTemplate):
    def __init__(self, template, image, command, image_pull_policy, key,
                 sliced_output_artifact=None, sliced_input_artifact=None,
                 sum_var=None, concat_var=None, auto_loop_artifacts=None,
                 group_size=None, format="%d", post_script="",
                 tmp_root="/tmp"):
        name = template.name
        super().__init__(name="%s-init-artifact" % name, image=image,
                         command=command, image_pull_policy=image_pull_policy)
        self.origin = template
        self.key = key
        self.sliced_output_artifact = sliced_output_artifact \
            if sliced_output_artifact is not None else []
        self.sliced_input_artifact = sliced_input_artifact \
            if sliced_input_artifact is not None else []
        self.sum_var = sum_var
        self.concat_var = concat_var
        self.auto_loop_artifacts = auto_loop_artifacts
        self.template = template
        self.group_size = group_size
        self.format = format
        self.post_script = post_script
        self.tmp_root = tmp_root

        if self.key is not None:
            self.inputs.parameters["dflow_group_key"] = InputParameter()
            # For the case of reusing sliced steps, ensure that the output
            # artifacts are reused
            for name in self.sliced_output_artifact:
                self.outputs.artifacts[name] = OutputArtifact(
                    path="%s/outputs/artifacts/%s" % (self.tmp_root, name),
                    save=S3Artifact(key="{{workflow.name}}/{{inputs."
                                    "parameters.dflow_group_key}}/%s" % name),
                    archive=None)
            self.outputs.parameters["dflow_artifact_key"] = OutputParameter(
                value="{{workflow.name}}/"
                "{{inputs.parameters.dflow_group_key}}")
        else:
            self.outputs.parameters["dflow_artifact_key"] = OutputParameter(
                value="{{workflow.name}}/{{pod.name}}")
            for name in self.sliced_output_artifact:
                self.outputs.artifacts[name] = OutputArtifact(
                    path="%s/outputs/artifacts/%s" % (self.tmp_root, name),
                    save=S3Artifact(key="{{workflow.name}}/{{pod.name}}/%s"
                                    % name),
                    archive=None)

        if self.sliced_input_artifact:
            if self.key is not None:
                self.inputs.parameters["dflow_key"] = InputParameter()
                if config["save_keys_in_global_outputs"]:
                    self.outputs.parameters["dflow_global"] = OutputParameter(
                        value="{{pod.name}}",
                        global_name="{{inputs.parameters.dflow_key}}",
                    )
            if group_size is not None:
                for name in self.sliced_input_artifact:
                    self.inputs.artifacts[name] = InputArtifact(
                        path="%s/inputs/artifacts/%s" % (self.tmp_root, name),
                        optional=True)
                    self.outputs.artifacts[name] = OutputArtifact(
                        path="%s/outputs/artifacts/%s" % (self.tmp_root, name),
                        archive=None, optional=True)
                self.outputs.parameters["dflow_ngroups"] = OutputParameter(
                    value_from_path="%s/outputs/parameters/dflow_ngroups"
                    % self.tmp_root, type=int)
            else:
                for name in self.sliced_input_artifact:
                    self.inputs.artifacts[name] = InputArtifact(
                        path="%s/inputs/artifacts/%s" % (self.tmp_root, name),
                        optional=True, sub_path=config["catalog_dir_name"])
                    self.outputs.parameters["dflow_slices_path"] = \
                        OutputParameter(
                            value_from_path="%s/outputs/parameters/dflow_"\
                                "slices_path" % self.tmp_root, type=dict)

        if self.sum_var is not None:
            name = self.sum_var.name
            self.inputs.parameters[name] = InputParameter()
            self.outputs.parameters["sum_%s" % name] = OutputParameter(
                value_from_path="%s/outputs/parameters/sum_%s" %
                (self.tmp_root, name), type=int)

        if self.concat_var is not None:
            name = self.concat_var.name
            self.inputs.parameters[name] = InputParameter()
            self.outputs.parameters["concat_%s" % name] = OutputParameter(
                value_from_path="%s/outputs/parameters/concat_%s" %
                (self.tmp_root, name), type=list)

        if self.auto_loop_artifacts:
            python_packages = []
            python_packages += __path__
            python_packages += jsonpickle.__path__
            self.python_packages = set(python_packages)
            self.inputs.artifacts["dflow_python_packages"] = InputArtifact(
                path="%s/inputs/artifacts/dflow_python_packages"
                % self.tmp_root)
            for name in self.origin.inputs.artifacts:
                if name in self.auto_loop_artifacts or (name.startswith(
                        "dflow_") and name[6:name.rfind("_")] in
                        self.auto_loop_artifacts):
                    self.inputs.artifacts[name] = InputArtifact(
                        path="%s/inputs/artifacts/%s" % (self.tmp_root, name),
                        optional=True)
            self.outputs.parameters["dflow_nslices"] = OutputParameter(
                value_from_path="%s/outputs/parameters/dflow_nslices" %
                self.tmp_root, type=int)

        self.render_script()

    def render_script(self):
        script = "import os, json\n"
        for name in self.sliced_output_artifact:
            script += "os.makedirs(r'%s/outputs/artifacts/%s/%s', "\
                "exist_ok=True)\n" % (self.tmp_root, name,
                                      config["catalog_dir_name"])
            script += "with open(r'%s/outputs/artifacts/%s/%s/init',"\
                " 'w') as f:\n" % (self.tmp_root, name,
                                   config["catalog_dir_name"])
            script += "    json.dump({'path_list': []}, f)\n"

        if self.sliced_input_artifact:
            required = []
            for i, name in enumerate(self.sliced_input_artifact):
                script += "path_list_%s = []\n" % i
                if self.group_size is not None:
                    script += "path = r'%s/inputs/artifacts/%s/%s'\n" % \
                        (self.tmp_root, name, config["catalog_dir_name"])
                else:
                    script += "path = r'%s/inputs/artifacts/%s'\n" % \
                        (self.tmp_root, name)
                script += "if os.path.exists(path):\n"
                script += "    for f in os.listdir(path):\n"
                script += "        with open(os.path.join(path, f), 'r')"\
                    " as fd:\n"
                script += "            for i in json.load(fd)['path_list']:\n"
                script += "                if i not in path_list_%s:\n" % i
                script += "                    path_list_%s.append(i)\n" % i
                script += "path_list_%s.sort(key=lambda x: x['order'])\n" \
                    % i
                if not self.origin.inputs.artifacts[name].optional:
                    required.append(i)

            if len(required) > 1:
                script += "assert " + " == ".join(
                    ["len(path_list_%s)" % i for i in required]) + "\n"

            if self.group_size is not None:
                script += "import math, shutil\n"
                script += "nslices = len(path_list_%s)\n" % required[0]
                script += "ngroups = math.ceil(nslices/%s)\n" % self.group_size
                if self.template.slices.shuffle:
                    script += "import random\n"
                    script += "random.seed(%s)\n" % \
                        self.template.slices.random_seed
                    script += "shuffled = list(range(nslices))\n"
                    script += "random.shuffle(shuffled)\n"
                    script += "random.seed()\n"  # unset seed
                script += "for i in range(ngroups):\n"
                for i, name in enumerate(self.sliced_input_artifact):
                    script += "    if not path_list_%s:\n" % i
                    script += "        continue\n"
                    script += "    group_dir = r'%s/outputs/artifacts/%s/"\
                        "group_' + ('%s' %% i)\n" % (
                            self.tmp_root, name, self.format)
                    script += "    os.makedirs(group_dir, exist_ok=True)\n"
                    if self.template.slices.shuffle:
                        script += "    path_list = [path_list_%s[j] for j in "\
                            "shuffled[%s*i:%s*(i+1)]]\n" % (
                                i, self.group_size, self.group_size)
                    else:
                        script += "    path_list = path_list_%s[%s*i:%s*(i+1)"\
                            "]\n" % (i, self.group_size, self.group_size)
                    script += "    for item in path_list:\n"
                    script += "        src = r'%s/inputs/artifacts/%s/'+item"\
                        "['dflow_list_item']\n" % (self.tmp_root, name)
                    script += "        dst = group_dir+'/'+item['dflow_list"\
                        "_item']\n"
                    script += "        os.makedirs(os.path.dirname(dst),"\
                        " exist_ok=True)\n"
                    script += "        if os.path.isfile(src):\n"
                    script += "            shutil.copy(src, dst)\n"
                    script += "        elif os.path.isdir(src):\n"
                    script += "            shutil.copytree(src, dst)\n"
                    script += "    cat = os.path.join(group_dir, '%s')\n" % \
                        config["catalog_dir_name"]
                    script += "    os.makedirs(cat, exist_ok=True)\n"
                    script += "    with open(os.path.join(cat, 'catalog'), "\
                        "'w') as f:\n"
                    script += "        json.dump({'path_list': path_list}, f)"\
                        "\n"
                script += "os.makedirs(r'%s/outputs/parameters', exist_ok="\
                    "True)\n" % self.tmp_root
                script += "with open(r'%s/outputs/parameters/dflow_ngroups'"\
                    ", 'w') as f:\n" % self.tmp_root
                script += "    f.write(str(ngroups))\n"
            else:
                script += "slices_path = []\n"
                script += "for i in range(len(path_list_%s)):\n" % required[0]
                script += "    item = {'order': i}\n"
                for i, name in enumerate(self.sliced_input_artifact):
                    script += "    item['%s'] = path_list_%s[i]"\
                        "['dflow_list_item'] if path_list_%s else None\n" % (
                            name, i, i)
                script += "    slices_path.append(item)\n"
                script += "os.makedirs(r'%s/outputs/parameters', exist_ok="\
                    "True)\n" % self.tmp_root
                script += "with open(r'%s/outputs/parameters/dflow_slices_"\
                    "path', 'w') as f:\n" % self.tmp_root
                script += "    json.dump(slices_path, f)\n"

        if self.sum_var is not None:
            name = self.sum_var.name
            script += """value = r'{{inputs.parameters.%s}}'
if "dflow_list_item" in value:
    dflow_list = []
    for item in json.loads(value):
        dflow_list += json.loads(item)
    var = list(map(lambda x: x['dflow_list_item'], dflow_list))
else:
    var = json.loads(value)
os.makedirs(r'%s/outputs/parameters', exist_ok=True)
with open(r'%s/outputs/parameters/sum_%s', 'w') as f:
    f.write(str(sum(map(int, var))))\n""" % (
                name, self.tmp_root, self.tmp_root, name)

        if self.concat_var is not None:
            name = self.concat_var.name
            script += """value = r'{{inputs.parameters.%s}}'
var = []
if "dflow_list_item" in value:
    for item in json.loads(value):
        for i in json.loads(item):
            var += i['dflow_list_item']
else:
    for item in json.loads(value):
        if isinstance(item, str):
            var += json.loads(item)
        else:
            var += item
os.makedirs(r'%s/outputs/parameters', exist_ok=True)
with open(r'%s/outputs/parameters/concat_%s', 'w') as f:
    f.write(json.dumps(var))\n""" % (
                name, self.tmp_root, self.tmp_root, name)

        if self.auto_loop_artifacts:
            from .python.python_op_template import handle_packages_script
            script += handle_packages_script(
                "%s/inputs/artifacts/dflow_python_packages" % self.tmp_root)
            script += "from dflow.python import Artifact\n"
            script += "from dflow.python.utils import handle_input_artifact\n"
            script += "from pathlib import Path\n"
            script += "from typing import List\n"
            script += "input = {}\n"
            required = []
            for name in self.auto_loop_artifacts:
                script += "input['%s'] = handle_input_artifact('%s', "\
                    "Artifact(List[Path]), None, r'%s', None, n_parts=%s, "\
                    "keys_of_parts=%s, prefix=%s)\n" % (
                        name, name, self.tmp_root,
                        self.origin.n_parts.get(name, None),
                        self.origin.keys_of_parts.get(name, None),
                        self.origin.input_artifact_prefix.get(name, None))
                if not self.origin.input_sign[name].optional:
                    required.append(name)

            if len(required) > 1:
                script += "assert " + " == ".join(
                    ["len(input['%s'])" % name for name in required]) + "\n"

            script += "os.makedirs(r'%s/outputs/parameters', "\
                "exist_ok=True)\n" % self.tmp_root
            script += "with open(r'%s/outputs/parameters/dflow_nslices', 'w')"\
                " as f:\n" % self.tmp_root
            script += "    f.write(str(len(input['%s'])))\n" % required[0]

        script += self.post_script.format(**{"tmp_root": self.tmp_root})
        self.script = script


class CheckNumSuccess(ShellOPTemplate):
    def __init__(self, name="check-num-success", image=None,
                 image_pull_policy=None):
        super().__init__(name=name, image=image,
                         image_pull_policy=image_pull_policy)
        self.command = ["python3"]
        self.script = "import json\n"
        self.script += "succ = '''{{inputs.parameters.success}}'''\n"
        self.script += "try:\n"
        self.script += "    succ = sum([int(i) for i in json.loads(succ)])\n"
        self.script += "except:\n"
        self.script += "    succ = 0\n"
        self.script += "print('Num success: ' + str(succ))\n"
        self.script += "print('Threshold: {{inputs.parameters.threshold}}')\n"
        self.script += "assert succ >= {{inputs.parameters.threshold}}\n"
        self.inputs = Inputs(
            parameters={"success": InputParameter(),
                        "threshold": InputParameter()})


class CheckSuccessRatio(ShellOPTemplate):
    def __init__(self, name="check-success-ratio", image=None,
                 image_pull_policy=None):
        super().__init__(name=name, image=image,
                         image_pull_policy=image_pull_policy)
        self.command = ["python3"]
        self.script = "import json\n"
        self.script += "succ = '''{{inputs.parameters.success}}'''\n"
        self.script += "try:\n"
        self.script += "    succ = sum([int(i) for i in json.loads(succ)])\n"
        self.script += "except:\n"
        self.script += "    succ = 0\n"
        self.script += "threshold = {{inputs.parameters.threshold}} * "\
            "{{inputs.parameters.total}}\n"
        self.script += "print('Num success: ' + str(succ))\n"
        self.script += "print('Threshold: ' + str(threshold))\n"
        self.script += "assert succ >= threshold\n"
        self.inputs = Inputs(
            parameters={"success": InputParameter(),
                        "total": InputParameter(),
                        "threshold": InputParameter()})



import abc
import base64
import functools
import inspect
import json
import logging
import os
import sys
from abc import ABC
from copy import deepcopy
from functools import partial
from importlib import import_module
from pathlib import Path
from typing import Dict, List, Set, Union

from ..argo_objects import ArgoObjectDict
from ..config import config
from ..context_syntax import GLOBAL_CONTEXT
from ..io import (InputArtifact, InputParameter, OutputArtifact,
                  OutputParameter, type_to_str)
from ..utils import dict2list, get_key, randstr, s3_config
from .vendor.typeguard import check_type
from .opio import OPIO, Artifact, BigParameter, OPIOSign, Parameter

iwd = os.getcwd()


def get_source_code(o):
    source_lines, start_line = inspect.getsourcelines(o)
    if sys.version_info.minor >= 9:
        source_file = inspect.getsourcefile(o)
    else:
        source_file = os.path.join(iwd, inspect.getsourcefile(o))
    with open(source_file, "r", encoding="utf-8") as fd:
        pre_lines = fd.readlines()[:start_line-1]
    return "".join(pre_lines + source_lines) + "\n"


class OP(ABC):
    """
    Python class OP

    Args:
        progress_total: an int representing total progress
        progress_current: an int representing currenet progress
    """
    progress_total = 1
    progress_current = 0
    key = None
    workflow_name = None

    def __init__(
            self,
            *args,
            **kwargs,
    ) -> None:
        pass

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        if key in ["progress_total", "progress_current"]:
            with open(os.environ.get("ARGO_PROGRESS_FILE",
                                     "ARGO_PROGRESS_FILE"), "w") as f:
                f.write("%s/%s" % (self.progress_current, self.progress_total))

    @classmethod
    @abc.abstractmethod
    def get_input_sign(cls) -> OPIOSign:
        """Get the signature of the inputs
        """

    @classmethod
    @abc.abstractmethod
    def get_output_sign(cls) -> OPIOSign:
        """Get the signature of the outputs
        """

    def _get_s3_link(self, key):
        if key[-4:] != ".tgz":
            key += "/"
        encoded_key = base64.b64encode(key.encode()).decode()
        return "%s/buckets/%s/browse/%s" % (
            s3_config["console"], s3_config["bucket_name"], encoded_key)

    def get_input_artifact_storage_key(self, name: str) -> str:
        templ = json.loads(os.environ.get("ARGO_TEMPLATE"))
        art = next(filter(lambda x: x["name"] == name,
                          templ["inputs"]["artifacts"]))
        return get_key(ArgoObjectDict(art))

    def get_input_artifact_link(self, name: str) -> str:
        key = self.get_input_artifact_storage_key(name)
        return self._get_s3_link(key)

    def get_output_artifact_storage_key(self, name: str) -> str:
        templ = json.loads(os.environ.get("ARGO_TEMPLATE"))
        art = next(filter(lambda x: x["name"] == name,
                          templ["outputs"]["artifacts"]))
        key = get_key(ArgoObjectDict(art), raise_error=False)
        if key is not None:
            return key

        key = get_key(ArgoObjectDict(templ["archiveLocation"]))
        if "archive" in art and "none" in art["archive"]:
            return "%s/%s" % (key, name)
        else:
            return "%s/%s.tgz" % (key, name)

    def get_output_artifact_link(self, name: str) -> str:
        key = self.get_output_artifact_storage_key(name)
        return self._get_s3_link(key)

    def register_output_artifact(self, name, namespace, dataset_name,
                                 **kwargs):
        if config["lineage"]:
            uri = self.get_output_artifact_storage_key(name)
            config["lineage"].register_artifact(
                namespace=namespace, name=dataset_name, uri=uri, **kwargs)
        else:
            logging.warning("Lineage client not provided")

    @abc.abstractmethod
    def execute(
            self,
            op_in: OPIO,
    ) -> OPIO:
        """Run the OP
        """
        raise NotImplementedError

    @staticmethod
    def exec_sign_check(func):
        @functools.wraps(func)
        def wrapper_exec(self, op_in):
            OP._check_signature(op_in, self.get_input_sign(), True)
            op_out = func(self, op_in)
            OP._check_signature(op_out, self.get_output_sign(), False)
            return op_out

        return wrapper_exec

    @staticmethod
    def _check_signature(
            opio: OPIO,
            sign: OPIOSign,
            is_input: bool,
    ) -> None:
        for ii in sign.keys():
            if ii not in opio.keys():
                if isinstance(sign[ii], Parameter) and hasattr(sign[ii],
                                                               "default"):
                    opio[ii] = sign[ii].default
                elif isinstance(sign[ii], BigParameter) and hasattr(
                        sign[ii], "default"):
                    opio[ii] = sign[ii].default
                elif isinstance(sign[ii], Artifact) and sign[ii].optional:
                    opio[ii] = None
                else:
                    if is_input:
                        raise RuntimeError('key %s declared in the input sign'
                                           ' is not present in the input'
                                           ' passed to the OP' % ii)
                    else:
                        raise RuntimeError('key %s declared in the output sign'
                                           ' is not present in the output'
                                           ' given by the OP' % ii)
        for ii in opio.keys():
            if ii not in sign.keys():
                if is_input:
                    raise RuntimeError(
                        'key %s in the input passed to the OP is not declared'
                        ' in its input sign' % ii)
                else:
                    raise RuntimeError(
                        'key %s in the output given by the OP is not declared'
                        ' in its output sign' % ii)
            io = opio[ii]
            ss = sign[ii]
            if isinstance(ss, Artifact):
                ss = ss.type
                if ss == Dict[str, str]:
                    ss = Dict[str, Union[str, None]]
                elif ss == Dict[str, Path]:
                    ss = Dict[str, Union[Path, None]]
                elif ss == List[str]:
                    ss = List[Union[str, None]]
                elif ss == List[Path]:
                    ss = List[Union[Path, None]]
                elif ss == Set[str]:
                    ss = Set[Union[str, None]]
                elif ss == Set[Path]:
                    ss = Set[Union[Path, None]]
            if isinstance(ss, Parameter):
                ss = ss.type
            # skip type checking if the variable is None
            if io is not None:
                check_type(ii, io, ss)

    @classmethod
    def function(cls, func=None, **kwargs):
        if func is None:
            return partial(cls.function, **kwargs)

        signature = func.__annotations__
        return_type = signature.get('return', None)
        input_sign = OPIOSign(
            {k: v for k, v in signature.items() if k != 'return'})
        output_sign, ret2opio, opio2ret = type2opiosign(return_type)

        class subclass(cls):
            task_kwargs = {}

            @classmethod
            def get_input_sign(cls):
                return deepcopy(input_sign)

            @classmethod
            def get_output_sign(cls):
                return deepcopy(output_sign)

            @OP.exec_sign_check
            def execute(self, op_in):
                op_out = func(**op_in)
                return ret2opio(op_out)

            def use(self, **kwargs):
                op = deepcopy(self)
                op.task_kwargs = kwargs
                return op

            def __call__(self, *args, **op_in):
                input_sign = self.get_input_sign()
                for i, v in enumerate(args):
                    k = list(input_sign)[i]
                    if k in op_in:
                        raise TypeError("%s() got multiple values for argument"
                                        " '%s'" % (func.__name__, k))
                    op_in[k] = v
                if GLOBAL_CONTEXT.in_context:
                    from ..task import Task
                    from .python_op_template import PythonOPTemplate
                    parameters = {k: v for k, v in op_in.items()
                                  if not isinstance(input_sign[k], Artifact)}
                    artifacts = {k: v for k, v in op_in.items()
                                 if isinstance(input_sign[k], Artifact)}
                    name = func.__name__.lower().replace("_", "-") + "-" + \
                        randstr()
                    task = Task(name,
                                template=PythonOPTemplate(self, **kwargs),
                                parameters=parameters, artifacts=artifacts,
                                **self.task_kwargs)
                    op_out = {**task.outputs.parameters,
                              **task.outputs.artifacts}
                    return opio2ret(op_out)
                return self.execute(op_in)

        subclass.func = func
        subclass.__name__ = func.__name__
        subclass.__module__ = func.__module__
        return subclass()

    @classmethod
    def superfunction(cls, func=None, **kwargs):
        if func is None:
            return partial(cls.superfunction, **kwargs)

        signature = func.__annotations__
        return_type = signature.get('return', None)
        input_sign = OPIOSign(
            {k: v for k, v in signature.items() if k != 'return'})
        output_sign, ret2opio, opio2ret = type2opiosign(return_type)

        from ..dag import DAG
        from ..task import Task

        class subclass(DAG):
            task_kwargs = {}

            def use(self, **kwargs):
                op = deepcopy(self)
                op.task_kwargs = kwargs
                return op

            def __call__(self, *args, **op_in):
                for i, v in enumerate(args):
                    k = list(input_sign)[i]
                    if k in op_in:
                        raise TypeError("%s() got multiple values for argument"
                                        " '%s'" % (func.__name__, k))
                    op_in[k] = v
                assert GLOBAL_CONTEXT.in_context
                parameters = {k: op_in[k] for k in self.inputs.parameters
                              if k in op_in}
                artifacts = {k: op_in[k] for k in self.inputs.artifacts
                             if k in op_in}
                name = self.name + "-" + randstr()
                task = Task(name, template=self, parameters=parameters,
                            artifacts=artifacts, **self.task_kwargs)
                op_out = {**task.outputs.parameters, **task.outputs.artifacts}
                return opio2ret(op_out)

        name = func.__name__.lower().replace("_", "-")
        with subclass(name=name, **kwargs) as dag:
            for n, s in input_sign.items():
                if isinstance(s, Artifact):
                    dag.inputs.artifacts[n] = InputArtifact(
                        type=s.type, optional=s.optional)
                else:
                    kw = {}
                    if isinstance(s, (Parameter, BigParameter)):
                        kw["type"] = s.type
                        if hasattr(s, "default"):
                            kw["value"] = s.default
                        if isinstance(s, BigParameter):
                            kw["save_as_artifact"] = (
                                config["mode"] != "debug")
                    else:
                        kw["type"] = s
                    dag.inputs.parameters[n] = InputParameter(**kw)
            outputs = func(**dag.inputs.parameters, **dag.inputs.artifacts)
            outputs = ret2opio(outputs)
            for n, s in output_sign.items():
                if isinstance(s, Artifact):
                    dag.outputs.artifacts[n] = OutputArtifact(
                        type=s.type, _from=outputs.get(n))
                else:
                    kw = {}
                    if isinstance(s, (Parameter, BigParameter)):
                        kw["type"] = s.type
                        if hasattr(s, "default"):
                            kw["default"] = s.default
                        if isinstance(s, BigParameter):
                            kw["save_as_artifact"] = (
                                config["mode"] != "debug")
                    else:
                        kw["type"] = s
                    kw["value_from_parameter"] = outputs.get(n)
                    dag.outputs.parameters[n] = OutputParameter(**kw)
        return dag

    @classmethod
    def get_opio_info(cls, opio_sign):
        opio = {}
        for io, sign in opio_sign.items():
            if type(sign) in [Artifact, Parameter, BigParameter]:
                opio[io] = sign
            else:
                opio[io] = type_to_str(sign)
        return opio

    @classmethod
    def get_info(cls):
        res = {}
        name = "%s.%s" % (cls.__module__, cls.__name__)
        res["name"] = name
        res["inputs"] = {k: str(v) for k, v in cls.get_opio_info(
            cls.get_input_sign()).items()}
        res["outputs"] = {k: str(v) for k, v in cls.get_opio_info(
            cls.get_output_sign()).items()}
        if hasattr(cls, "func"):
            res["execute"] = "".join(inspect.getsourcelines(cls.func)[0])
        else:
            res["execute"] = "".join(inspect.getsourcelines(cls.execute)[0])
        return res

    @classmethod
    def convert_to_graph(cls):
        source = None
        try:
            if hasattr(cls, "_source"):
                mod = "__main__"
                source = cls._source
            else:
                mod = str(cls.__module__)
                source = get_source_code(cls.func) if hasattr(
                    cls, "func") else get_source_code(cls)
        except Exception:
            logging.info("Failed to get source code of OP", exc_info=True)
        return {
            "module": mod,
            "name": str(cls.__name__),
            "inputs": cls.get_opio_info(cls.get_input_sign()),
            "outputs": cls.get_opio_info(cls.get_output_sign()),
            "source": source,
        }

    @classmethod
    def from_graph(cls, graph):
        if graph["module"] in ["__main__", "__mp_main__"]:
            exec(graph["source"], globals())
            op = globals()[graph["name"]]
            if isinstance(op, OP):
                op.__class__._source = graph["source"]
            else:
                op._source = graph["source"]
        else:
            mod = import_module(graph["module"])
            op = getattr(mod, graph["name"])
        return op


def type2opiosign(t):
    from typing import Tuple
    try:
        from typing import _GenericAlias as TupleMeta
    except ImportError:
        from typing import TupleMeta
    if isinstance(t, dict):
        return OPIOSign({k: v for k, v in t.items()}), lambda x: x, lambda x: x
    elif hasattr(t, "__annotations__") and issubclass(t, dict):
        return OPIOSign(
            {k: v for k, v in t.__annotations__.items()}), lambda x: x, \
            lambda x: x
    elif hasattr(t, "__annotations__") and issubclass(t, tuple):
        return OPIOSign(
            {k: v for k, v in t.__annotations__.items()}), \
            lambda x: {k: v for k, v in zip(t.__annotations__, x)}, \
            lambda x: tuple(dict2list({list(t.__annotations__).index(k): v
                                       for k, v in x.items()}))
    elif isinstance(t, TupleMeta) and t.__origin__ in [tuple, Tuple]:
        return OPIOSign(
            {"dflow_output_%s" % i: v for i, v in enumerate(t.__args__)}), \
            lambda x: {"dflow_output_%s" % i: v for i, v in enumerate(x)}, \
            lambda x: tuple(dict2list({int(k[13:] if k.startswith(
                "dflow_output_") else k): v for k, v in x.items()}))
    elif t is not None:
        return OPIOSign({"dflow_output": t}), lambda x: {"dflow_output": x}, \
            lambda x: x["dflow_output"]
    else:
        logging.warning(
            'We recommended using return type signature like:'
            '\n'
            "def func()->TypedDict('op', {'x': int, 'y': str})"
            '\nor\n'
            "def func()->NamedTuple('op', [('x', int), ('y', str)])")
        return OPIOSign({}), lambda x: {}, lambda x: None
